module LikelihoodDegenerations

using CapAndHomalg
using Combinatorics
using HomotopyContinuation
using LinearAlgebra

import HomotopyContinuation
import CapAndHomalg

function __init__()
	LoadPackage( "MatroidGeneration" )
	LoadPackage( "ZariskiFrames" )
	LoadPackage( "Sheaves" )
	#CapAndHomalg.SetInfoLevel( InfoWarning, 0 )
	global QQ = GAP.Globals.HomalgFieldOfRationalsInSingular()
end

export QQ, matroid_from_database, matroid_from_revlex_string, realization_space
export RealizedMatroidFromDatabase, solve_ml_equations

export maximum_likelihood_equations, ml_degree, find_valuations, compute_soft_limit_valuations

export X
#export RealizedMatroid, test_all_matroids_for_realization, matroid_from_database, matroid_from_revlex_string, QQ


function revLexOrder(f1::AbstractVector{Int64}, f2::AbstractVector{Int64})::Bool
	if maximum( f1 ) < maximum( f2 )
    	return true
    elseif maximum( f1 ) > maximum( f2 )
    	return false
	else
    	if length( f1 ) > 1
      		return revLexOrder( view(f1,1:(length( f1 )-1)), view(f2,1:(length( f2 )-1)))
      	else
      		return false
      	end
    end
end

function matroid_from_revlex_string( n, rank, revlex_string )
	allBases = collect(combinations( 1:n, rank ))
	Base.sort!(allBases;lt=revLexOrder)

	bases =  [GapObj(allBases[i]) for i in 1:length(allBases) if revlex_string[i]=='*']

    return GAP.Globals.MatroidByBases( n, GapObj(bases) )
end

#Creates a GAP matroid from the rank, n_elts, and index of a matroid
#  in the database
function matroid_from_database(n::Int,r::Int,line::Int)
	FileName = "MatroidDB/SimpleMatroid"*string(r)*string(n)*".txt"
	s=readlines(FileName)[line]
	println(s)
	return matroid_from_revlex_string(n,r,s)
end

#Turns a matroid into a realization space in GAP.
function realization_space( matroid )
	M = GAP.Globals.EquationsAndInequationsOfModuliSpaceOfMatroid( matroid, QQ )
	m = GAP.Globals.DistinguishedQuasiAffineSet( M )
	e = GAP.Globals.EmbedInSmallerAmbientSpace( m )
	if string(e) == string(m)
		return m
	else
		return e
	end
end

function pull_equations(E)
	startingpos=findfirst(x->x=='(',E)
	endingpos=findfirst(x->x==')',E)
	if startingpos == nothing
		return "[]"
	else
		return "["*E[startingpos+1:endingpos-1]*"]"
	end
end

function pull_variables(s)
	if s=="Q"
		return nothing
	end
	BrokenUp=[]
	CurrentRead=""
	Started=false
	finished=false
	position=1
	while finished==false
		if s[position]=='['
			position+=1
			CurrentRead=""
			Started=true
		end
		if s[position]==','
			push!(BrokenUp,CurrentRead)
			CurrentRead=""
		elseif s[position]==']'
			push!(BrokenUp,CurrentRead)
			finished=true
		else
			CurrentRead=CurrentRead*s[position]
		end
		position+=1
	end
	Command="@var "
	for b in BrokenUp
		Command=Command*b*","
	end
	Command=Command[1:end-1]
	return Command
end

mutable struct RealizedMatroid
	parametrization
	variables
	equations
	bases
	nonbases
	dimension
	degree
	mldegree
	variable_dictionary
end

function Base.show(io::IO, S::RealizedMatroid)
	if S.dimension == -1
		println(io,"Non realizable matroid")
		println(io,"\n with "*string(length(S.nonbases))*" vanishing Plucker coordinate(s)")
	elseif S.dimension == 0
		println(io,"Zero dimensional realized matroid")
		show(io,"text/plain",S.parametrization)
	else
    	println(io,"Realized matroid in dimensions "*string(size(S.parametrization)))
    	show(io,"text/plain",S.parametrization)
		println(io,"\nWith equations")
    	show(io,S.equations)
    	println(io,"\nand "*string(length(S.nonbases))*" vanishing Plucker coordinate(s)")
	end
end


function RealizedMatroidFromDatabase(r::Int,n::Int,line::Int)
	M = matroid_from_database(n,r,line)
	e = realization_space(M)

	dimension = GAP.Globals.Dimension(e)
	degree = GAP.Globals.Degree(e)

	allCombinations = collect(combinations(1:GAP.Globals.Size(M),GAP.Globals.Rank(M)))
	bases = Vector{Vector{Int64}}(gap_to_julia(GAP.Globals.Bases(M)))
	nonbases = setdiff(allCombinations,bases)

#	number_of_components = GAP.Globals.Length(GAP.Globals.IrreducibleComponents(GAP.Globals.Closure(e)))


	if dimension == -1
		return RealizedMatroid(nothing,nothing,nothing,bases,nonbases,-1,nothing,nothing,nothing)
	end

	matrix = string(GAP.Globals.EntriesOfHomalgMatrixAsListList(GAP.Globals.MatrixOfVectorMatroid(GAP.Globals.ParametrizedObject(e))))




	A = GAP.Globals.DistinguishedLocallyClosedPart( e )
	R = GAP.Globals.UnderlyingRing( GAP.Globals.Closure( e ) )

	equations = ""
	if GAP.Globals.IsObjectInMeetSemilatticeOfMultipleDifferences( A )
		equations = string(R / GAP.Globals.UnderlyingMatrix( GAP.Globals.MorphismOfUnderlyingCategory( A.I ) ))
	else
		equations = string(R / GAP.Globals.UnderlyingMatrix( GAP.Globals.MorphismOfUnderlyingCategory( A ) ))
	end

	PV=pull_variables(equations)
	if PV == nothing ##if there are no variables, it must be a point.
		if dimension == 0

			parametrization=convert(Array{Expression,2},hcat(eval(Meta.parse(matrix))...)')

			return RealizedMatroid(parametrization,nothing,nothing,bases,nonbases,dimension,degree,degree,nothing)
		else
			return "ERROR"
		end
	else PV != nothing
		VarString="["*PV[5:end]*"]"
		#This defines the variables so we can use them.
		eval(Meta.parse(PV))
		Vars=eval(Meta.parse(VarString))
		EquationString=pull_equations(equations)
		Equations = Array{Expression,1}(eval(Meta.parse(EquationString)))

		parametrization=convert(Array{Expression,2},hcat(eval(Meta.parse(matrix))...)')


		RealizedMatroid(parametrization,Vars,Equations,bases,nonbases,dimension,degree,nothing,nothing)
	end
end



##Constructing likelihood equations from realized matroid

function maximum_likelihood_equations(Z::RealizedMatroid)
    SP=maximum_likelihood_potential(Z)
    lagrange_equations(Z.equations,SP,Z.variables)
end

function maximum_likelihood_potential(Z::RealizedMatroid)
	#produces a variable for each basis
	s = [Variable(:s,ℓ...) for ℓ in Z.bases]
	#now we construct a dictionary from the basis index
	#  to these variables
    VarDict=Dict{typeof(Z.bases[1]),typeof(s[1])}()
    for i in 1:length(Z.bases)
        VarDict[Z.bases[i]]=s[i]
    end
    M=Z.parametrization
    S = sum([VarDict[I]*log(det(M[:,I])) for I in Z.bases])
	Z.variable_dictionary = VarDict
    return S
end

function gradient(S,x)
    System(
        vec(differentiate(S, [x...])),
        variables = [x...],
        parameters = setdiff(variables(S), [x...]),
    )
end

function lagrange_equations(G,f,Vars)
    M=length(G)
    @var λ[1:M]
    L = f
    if M>0
        L = f-sum([λ[i]*G[i] for i in 1:M])
    end
    λx = vcat(vec(λ),vec(Vars))
    ∇L = gradient(L,λx)
    return ∇L
end



####Numerical algorithms for producing points on each component
#### of a realization space. Or for finding start points for
#### monodromy. Or for solving ML equations.

#returns a witness point on each irreducible component
function matroid_witness_point(Z::RealizedMatroid)
	F=System(Z.equations,variables=Z.variables)
	if length(Z.equations)==0
		println("Completely parametrized. Returning a single point.")
		return [randn(ComplexF64,length(Z.variables))]
	end
	W=witness_set(F,codim=length(Z.variables)-Z.dimension, compile=false)
	ishom=false
	if codim(W) != length(Z.variables)-Z.dimension
		ishom=true
	end
	BU = monodromy_breakup(W; ishom=ishom)
	#if hom, then rescale so it's actually random

	Representatives=[b[1] for b in BU]
	if ishom==true
		randcomp=rand(ComplexF64)
		Representatives=[randcomp*r for r in Representatives]
	end
	return Representatives
end



###Given a witness set, this breaks it up into components
function monodromy_breakup(W; ishom=false)
	LeftToClassify=solutions(W)
	Buckets=[]
	F = try
		F=W.F.compiled.system
	catch
		W.F.system
	end
	P=W.L
	while length(LeftToClassify)>0
		println("$(length(LeftToClassify)) solutions left to classify")
		s=LeftToClassify[1]
		counter=0
		success=false
		MS=nothing
		while success==false && counter<5
			counter+=1
			MS=monodromy_solve(F,[s],P)
			println("Found $(length(solutions(MS))) solutions")
			if MS.returncode == :success
				success=true
			else
			end
		end
		MSsols=solutions(MS)
		if ishom==true
			LeftToClassify=[l/l[1] for l in LeftToClassify]
			MSsols=[l/l[1] for l in MSsols]
		end

		unfoundlocs=findall(x->true ∉ [isapprox(x,y;rtol=0.0001) for y in MSsols],LeftToClassify)
		if length(unfoundlocs)+length(MSsols)!=length(LeftToClassify)
			println("Something went wrong. Could not find the monodromy solutions in lefttoclassify")
		end
		push!(Buckets,MSsols)
		LeftToClassify=LeftToClassify[unfoundlocs]

	end

	return Buckets
end

function ml_degree(Z::RealizedMatroid)
	if Z.dimension == 0
		return [Z.degree]
	elseif Z.parametrization == nothing
		return "Not realizable"
	end
	MS_tuple = solve_ml_equations(Z)
	MLD_tuple = [length(solutions(m)) for m in MS_tuple]
	return MLD_tuple
end

function solve_ml_equations(Z::RealizedMatroid;S=nothing,P=nothing)
    F=maximum_likelihood_equations(Z)
    if S==nothing && P==nothing
		#Construct a start point for monodromy on critical equations
		ComponentWitnesses=matroid_witness_point(Z)
		MS_tuple=[]
		for s in ComponentWitnesses
			lambdaCount=length(F.variables)-length(Z.variables)
			VarPoint=vcat(ones(lambdaCount),s)
			EvaluatedPolys=evaluate(F.expressions[1:end],F.variables=>VarPoint)
			filter!(x->length(exponents_coefficients(x,F.parameters)[2])>1,EvaluatedPolys)
			A=linsystomatrix(EvaluatedPolys,F.parameters)
			N = nullspace(A)
			P = N*randn(ComplexF64,size(N,2))
			P = P/P[1]
			P = P[2:end]
			StartSol=vcat(ones(lambdaCount),s)
			StartParam=P
			M = solve_ml_equations(Z;S=StartSol,P=StartParam)
			push!(MS_tuple,M)
		end
        return MS_tuple
    else
		#first we move to something random in the parameters
		# instead of in the solutions
		randomTarget=randn(ComplexF64,length(P))
		newS = solutions(solve(F,S;start_parameters=P,target_parameters=randomTarget))
        M=monodromy_solve(F,newS,randomTarget)
        return M
    end
end

function linsystomatrix(linsys,vars)
	b=[evaluate(lineq,vars=>zeros(length(vars))) for lineq in linsys]
	newlinsys=[]
	for i in 1:length(b)
		push!(newlinsys,linsys[i]-b[i])
	end
    Id = Matrix(I,length(vars),length(vars))
    cfs = [ System([subs(lineq,vars=>Id[:,j][:])])([])[1] for lineq ∈ newlinsys, j = 1:length(vars)]
	cfs=hcat(b,cfs)
    convert(Array{ComplexF64,2},cfs)
end






##############For Certification of Matroid ML Degrees


function certify_all_mldegrees(k,n; ntries=10, range=0:0)
	if range==0:0
		N=length(readlines("MatroidDB/SimpleMatroid"*string(k)*string(n)*".txt"))
		range = 1:N
	end
	for i in range
		certify_mldegree_lowerbound(k,n,i;save_certificates=true,ntries=ntries)
	end
end

function certify_mldegree_lowerbound(k,n,i;save_certificates=false,ntries=1)
	Z = RealizedMatroidFromDatabase(k,n,i)

	#If zero dimensional
	if Z.dimension == 0
		if save_certificates==true
			OutString = ""
			OutString = OutString*"Matroid number "*string(i)*" of simple rank "*string(k)*" matroids on "*string(n)*" elements.\n"
			FileName = "MatroidDB/SimpleMatroid"*string(k)*string(n)*".txt"
			s=readlines(FileName)[i]
			OutString = OutString*s*"\n\n"
			OutString = OutString*"Best Certified ML Degree\n"*"Any"*string([Z.degree])*"\n\n"
			OutString = OutString*"All computed ML Degrees\n"
			for c in 1:ntries
				OutString = OutString*"Any"*string([Z.degree])*"\n"
			end
			println(OutString)
			if isdir("MatroidCertificates")==false
				mkdir("MatroidCertificates")
			end
			if isdir("MatroidCertificates/"*string(k)*string(n))==false
				mkdir("MatroidCertificates/"*string(k)*string(n))
			end

			open("MatroidCertificates/"*string(k)*string(n)*"/MatroidSummary_"*string(i), "w") do io
				write(io,OutString)
			end
			open("MatroidCertificates/"*string(k)*string(n)*"/Certificate_"*string(i), "w") do io
				write(io,"No certificate necessary. Calculation was symbolic.")
			end
		end
		println("Zero-dimensional. No certification necessary")
		return(Z)
	end
	##If not realizable
	if Z.dimension == -1
		if save_certificates==true
			OutString = ""
			OutString = OutString*"Matroid number "*string(i)*" of simple rank "*string(k)*" matroids on "*string(n)*" elements.\n"
			FileName = "MatroidDB/SimpleMatroid"*string(k)*string(n)*".txt"
			s=readlines(FileName)[i]
			OutString = OutString*s*"\n\n"
			OutString = OutString*"Not Realizable"
		end
		println("Not Realizable")
		return(Z)
	end
	MatroidDimension = Z.dimension
	MLDs_Computed=[]
	MLDs_Certified=[]
	BestCertifiedCount=[]
	BestCertificates=[]
	for trynum in 1:ntries
		println("Try number: ",trynum)
		ComponentWitnesses=matroid_witness_point(Z) #returns a point on each irr. comp of top dim of matroid
		println("  Number of components: ",length(ComponentWitnesses))
		MLD_bucket=[]
		MLD_cert_bucket=[]
		ComponentCertificates=[]
		for S in ComponentWitnesses #for each component, compute its ml degree
			MLEq=maximum_likelihood_equations(Z)
			#Construct a start point for monodromy on critical equations
			lambdaCount=length(MLEq.variables)-length(Z.variables)
			VarPoint=vcat(ones(lambdaCount),S)
			EvaluatedPolys=evaluate(MLEq.expressions[1:end],MLEq.variables=>VarPoint)
			filter!(x->length(exponents_coefficients(x,MLEq.parameters)[2])>1,EvaluatedPolys)
			A=linsystomatrix(EvaluatedPolys,MLEq.parameters)
			N = nullspace(A)
			P = N*randn(ComplexF64,size(N,2))
			P = P/P[1]
			P = P[2:end]
			StartSol=vcat(ones(lambdaCount),S)
			StartParam=P
			println("...Solving Critical Point Equations")
			#Solve the critical point equations
			MS = solve_ml_equations(Z;S=StartSol,P=StartParam)
			Sols=solutions(MS)
			println("  Found solutions: ",length(Sols))
			MSP=MS.parameters
			#If the number of solutions is smaller than the best certificate, there's no reason to certify
			#Only certify if it is larger than the minimum of the BestCertifiedCount (for irred. this just means larger, for red. [3,3] would be certified against [3,2])
			CertifiableSystem=nothing
			CertResults=nothing
			CertCount=0
			if BestCertifiedCount==[] || length(Sols)>min(BestCertifiedCount...)
				if length(Sols)==0
					println("This may have ML degree zero. Not able to certify.")
				else
				println("  This result is better than the best certified")
					CertifiableSystem = System(evaluate(MLEq.expressions[1:end],MLEq.parameters=>MSP),variables=MLEq.variables)
					CertResults = certify(CertifiableSystem,Sols)
					CertCount = length(distinct_certificates(CertResults))
				end
			end
			push!(ComponentCertificates,CertResults)
			push!(MLD_bucket,length(Sols))
			push!(MLD_cert_bucket,CertCount)
		end
		push!(MLDs_Computed,MLD_bucket)
		push!(MLDs_Certified,MLD_cert_bucket)
		#Check if a certificate was computed. If so compare it to the best certificate.
		#If ANY coordinate increases, replace the best certificate

		#if nothing is the best yet
		if BestCertifiedCount == []
			BestCertifiedCount=MLD_cert_bucket
			BestCertificates = ComponentCertificates
			println("  Best certificate updated")
		#if something has been certified
		elseif length(MLD_cert_bucket)>0
			if MLD_cert_bucket[1]>BestCertifiedCount[1] || (length(MLD_cert_bucket)>1 && MLD_cert_bucket[2]>BestCertifiedCount[2])
				BestCertifiedCount=MLD_cert_bucket
				BestCertificates = ComponentCertificates
				println("  Best certificate updated")
			end
		end
		println("\n\n")
	end
	OutString = ""
	OutString = OutString*"Matroid number "*string(i)*" of simple rank "*string(k)*" matroids on "*string(n)*" elements.\n"
	FileName = "MatroidDB/SimpleMatroid"*string(k)*string(n)*".txt"
	s=readlines(FileName)[i]
	OutString = OutString*s*"\n\n"
	OutString = OutString*"Best Certified ML Degree\n"*string(BestCertifiedCount)*"\n\n"
	OutString = OutString*"All computed ML Degrees\n"
	for c in MLDs_Computed
		OutString = OutString*string(c)*"\n"
	end
	println(OutString)


	if save_certificates==true
		if isdir("MatroidCertificates")==false
			mkdir("MatroidCertificates")
		end
		if isdir("MatroidCertificates/"*string(k)*string(n))==false
			mkdir("MatroidCertificates/"*string(k)*string(n))
		end

		open("MatroidCertificates/"*string(k)*string(n)*"/MatroidSummary_"*string(i), "w") do io
			write(io,OutString)
		end
		if length(BestCertificates)==1
			if BestCertificates[1] != nothing
				save("MatroidCertificates/"*string(k)*string(n)*"/Certificate_"*string(i),BestCertificates[1])
			elseif BestCertifiedCount==[0]
				open("MatroidCertificates/"*string(k)*string(n)*"/Certificate_"*string(i), "w") do io
					write(io,"No certificate necessary. Lower bound is zero. Numerically the ML degree is zero.")
				end
			end
		else
			save("MatroidCertificates/"*string(k)*string(n)*"/Certificate_"*string(i)*"_1",BestCertificates[1])
			save("MatroidCertificates/"*string(k)*string(n)*"/Certificate_"*string(i)*"_2",BestCertificates[2])
		end
	end
	println(MLDs_Computed)
	println(MLDs_Certified)
	println(BestCertifiedCount)
	println(BestCertificates)
end
########################################
#### For likelihood degenerations
########################################

function X(k,n)
	RealizedMatroidFromDatabase(k,n,1)
end

function soft_limit_equations(Z::RealizedMatroid)
	F = maximum_likelihood_equations(Z)
	(k,n) = size(Z.parametrization)
	indexSets = Z.bases
	last_particle_dependent = filter(ℓ -> maximum(ℓ)==n,indexSets)
	lpd_parameters = [Z.variable_dictionary[I] for I in last_particle_dependent]
	@var τ
	soft_limit_eqs = subs(F.expressions,lpd_parameters=>τ*lpd_parameters)
	soft_limit_system = System(soft_limit_eqs,variables = F.variables, parameters = vcat(F.parameters,τ))
	return soft_limit_system
end

function solve_ml_eqs_nostart(Z::RealizedMatroid)
	ml_eqs = maximum_likelihood_equations(Z)
	MonSolve = monodromy_solve(ml_eqs)
	Params = MonSolve.parameters
	Sols = solutions(MonSolve)
	return(Params,Sols)
end

#interpolates the data v with respect to the t-values target_params
# to approximate the valuations
function Rico(v,target_params)
    l = [log(tp[1]) for tp ∈ target_params]
    coeffmtx = [ones(length(l)) l]
    rhs = v
    x = coeffmtx\rhs
    #println(x[2])
	r=nothing
	##Things may be so numerically unstable that x is not a number.
	## This occurs if we tried to get too close to soft_limit=0
	## If that's the case, take fewer samples toward soft_limit->0
	## and try again.
	if string(x[2])!="NaN"
		r=round(Int,x[2])
	else
		r=Rico(v[1:length(v)-5],target_params[1:length(v)-5])
	end
    return r
end


# Takes a single soft_limit->0 sampling of a solution and
# computes the valuation.
function determine_valuation(Z,sollist,t_param_list)
	#take out bases which are constant
	non_constant_bases = filter(x-> (x ⊆ collect(1:size(Z.parametrization)[1]+1)) == false,Z.bases)
	t_param_list=t_param_list[1:length(sollist)]
	interpolmtx = fill(0.0,0,length(non_constant_bases))
	for sol ∈ sollist
		##MM is the solution as a matrix
		MM = evaluate.(Z.parametrization,vec(Z.variables) => sol[1:length(vec(Z.variables))])
		MM = Matrix{ComplexF64}(MM)
		interpolmtx = vcat(interpolmtx,transpose([log(abs(det(MM[:,ii]))) for ii in non_constant_bases]))
	end
	vals = [Rico(interpolmtx[:,i],t_param_list) for i = 1:length(non_constant_bases)]
	return(vals)
end

#Makes a dictionary with elements of S as keys and frequency of that element as
#  the corresponding value
function tally(S)
    D=Dict{Any,Int}()
    for s in S
        D[s]=get(D,s,0)+1
    end
    return D
end

function compute_soft_limit_valuations(Z::RealizedMatroid)
	println("Solving the likelihood equations...")
	(Params,Sols) = solve_ml_eqs_nostart(Z)
	T = find_valuations(Z,Params,Sols)
	return T
end

#Takes a subset of the solutions of the maximum likelihood equations of a
#  matroid stratum, embeds these solutions into the soft limit, and approximates
#  the valuations of the plucker coordinates by monitoring numerically how
#  that soft limit goes to zero. Returns a tally of valuations
function find_valuations(Z::RealizedMatroid,Params,Sols)
	println("Tracking each of the ", length(Sols), " solutions given along the
	soft limit toward zero.")
	flush(stdout)
	(sample_list,t_param_list) = sample_soft_near_zero(Z,Params,Sols)
	println("Numerically computing the valuation of each path from this data...")
	flush(stdout)
	valuation_list = []
	for i in 1:length(sample_list)
		if Int(floor(i/1000)) == i/1000
			println("Progress: at valuation number ",i)
			flush(stdout)
		end
		s=sample_list[i]
		push!(valuation_list,determine_valuation(Z,s,t_param_list))
	end
	valuation_list = [determine_valuation(Z,s,t_param_list) for s in sample_list]
	println("Tallying...")
	flush(stdout)
	T = tally(valuation_list)
	return T
end


#Takes a collection of solutions to the maximum likelihood equations and
# tracks each toward soft_limit -> 0. Outputs a list S of lists. where each
# element s of S corresponds to a single solution, and the elements of
# s correspond to each successive soft_limit value of that path.
function sample_soft_near_zero(Z::RealizedMatroid,Params,Sols)
	ml_eqs = maximum_likelihood_equations(Z)
	soft_start_params=vcat(Params,1)
	target_params_in_t = [10.0^(-k) for k=2:0.1:6]
	soft_target_params=[vcat(Params,t) for t in target_params_in_t]
	soft_limit_sys = soft_limit_equations(Z)
	samples = solve(soft_limit_sys,Sols,
		start_parameters = soft_start_params,
		target_parameters = soft_target_params)
	samples_per_solution=[]
	for i in 1:length(Sols)
		not_lost=true
		t_num=0
		single_solution_sample=[]
		while not_lost==true && t_num<length(target_params_in_t)
			t_num+=1
			#single_t_result is all the solutions at time target_parameters_in_t[t_num]
			i_path_result = samples[t_num][1].path_results[i]
			if i_path_result.return_code == :success
				push!(single_solution_sample,i_path_result.solution)
			else
				not_lost=false
			end
		end
		push!(samples_per_solution,single_solution_sample)
	end
	return(samples_per_solution,target_params_in_t)
end


#=

mutable struct RealizedMatroid
	matrix::String
	equations::String
	dimension::Int64
	degree::Int64
0	smoothness::Int64
	number_of_components::Int64
	nonbases::Vector{Vector{Int64}}
end
function RealizedMatroid(matroid)
	e = realization_space(matroid)
	dimension = Dimension(e)
	degree = Degree(e)

	allCombinations = collect(combinations(1:Size(matroid),Rank(matroid)))
    bases = Vector{Vector{Int64}}(gap_to_julia(Bases(matroid)))
    nonBases = setdiff(allCombinations,bases)

    number_of_components = Length(IrreducibleComponents(Closure(e)))

	if dimension == -1
		return RealizedMatroid("empty","empty", dimension, degree, number_of_components, nonBases)
	end

	matrix = string(GAP.Globals.EntriesOfHomalgMatrixAsListList(GAP.Globals.MatrixOfVectorMatroid(ParametrizedObject(e))))
	A = DistinguishedLocallyClosedPart( e )
	R = UnderlyingRing( Closure( e ) )

	if IsObjectInMeetSemilatticeOfMultipleDifferences( A ) then
        equations = string(R / UnderlyingMatrix( MorphismOfUnderlyingCategory( A.I ) ))
    else
        equations = string(R / UnderlyingMatrix( MorphismOfUnderlyingCategory( A ) ))
    end

	return RealizedMatroid(matrix, equations, dimension, degree, number_of_components, nonBases)
end

function RealizedMatroid(n,rank,revlex_string)
	matroid = matroid_from_revlex_string(n, rank, revlex_string)
	return RealizedMatroid(matroid)
end

function test_all_matroids_for_realization(n,r,bound)
	for i in 1:bound
		matroid = matroid_from_database(n,r,i)
		println(RealizedMatroid(matroid))
	end
end






=#
end # module
